@@ -9,20 +9,28 @@ from urllib.parse import urlparse
 from pydantic import BaseModel, BaseSettings
 
 from labmachine import defaults, utils
-from labmachine.base import DNSSpec, ProviderSpec
+from labmachine.base import ComputeSpec, DNSSpec
 from labmachine.io.kvspec import GenericKVSpec
 from labmachine.types import (AttachStorage, BlockStorage, BootDiskRequest,
                               DNSRecord, DNSZone, GPURequest, InstanceType,
                               Permissions, StorageRequest, VMInstance,
                               VMRequest)
 
-VM_PROVIDERS = {"gce": "labmachine.providers.google.compute.GCEProvider"}
+VM_PROVIDERS = {"gce": "labmachine.providers.google.Compute"}
 DNS_PROVIDERS = {
-    "gce": "labmachine.providers.google.dns.GoogleDNS",
+    "gce": "labmachine.providers.google.GoogleDNS",
     "cloudflare": "labmachine.providers.cloudflare.dns.CloudflareDNS",
 }
 
 
+def _check_dns_env_var():
+    _env = os.getenv(defaults.JUP_DNS_KEY)
+    if _env:
+        return defaults.JUP_DNS_KEY
+    else:
+        return defaults.GOOGLE_AUTH_ENV
+
+
 class JupyterInstance(BaseModel):
     container: str = "jupyter/minimal-notebook:python-3.10.6"
     uuid: str = "1000"
@@ -52,10 +60,17 @@ class JupyterVolume(BaseModel):
     labels: Dict[str, Any] = {},
 
 
+class ContainerRegistry(BaseModel):
+    name: str
+    project: str
+    location: str
+
+
 class JupyterConfig(BaseSettings):
     VOLUME: Optional[JupyterVolume]
     INSTANCE: JupyterInstance
     STATE_PATH: str = "state.json"
+    REGISTRY: Optional[ContainerRegistry] = None
 
     class Config:
         env_prefix = "JUP_"
@@ -72,6 +87,7 @@ class JupyterState(BaseModel):
     vm: Optional[VMInstance] = None
     url: Optional[str] = None
     record: Optional[DNSRecord] = None
+    registry: Optional[ContainerRegistry] = None
 
 
 class LabResponse(BaseModel):
@@ -80,7 +96,7 @@ class LabResponse(BaseModel):
     url: str
 
 
-def load_jupyter_conf(settings_module) -> JupyterConfig:
+def load_conf_module(settings_module) -> JupyterConfig:
     mod = import_module(settings_module)
 
     settings_dict = {}
@@ -145,7 +161,7 @@ def push_state(state: JupyterState) -> str:
     return fp
 
 
-def find_gce(prov: ProviderSpec, ram: int, cpu: str, gpu="") \
+def find_gce(prov: ComputeSpec, ram: int, cpu: str, gpu="") \
         -> List[InstanceType]:
     sizes = prov.driver.list_sizes(location=prov._location)
     node_types = []
@@ -172,31 +188,34 @@ def find_gce(prov: ProviderSpec, ram: int, cpu: str, gpu="") \
     return node_types
 
 
-def find_node_types(prov: ProviderSpec, ram=2, cpu=2, gpu="") \
+def find_node_types(prov: ComputeSpec, ram=2, cpu=2, gpu="") \
         -> List[InstanceType]:
-    if self.prov.providerid == "gce":
+    if prov.providerid == "gce":
         nodes = find_gce(prov, ram, cpu, gpu)
     return nodes
 
 
 class JupyterController:
 
-    def __init__(self, compute: ProviderSpec,
+    def __init__(self, compute: ComputeSpec,
                  dns: DNSSpec,
                  state: JupyterState
                  ):
-        self.prov = compute
+        self.compute = compute
         self.dns = dns
         self._zone = self.dns.get_zone(state.zone_id)
         self._state: JupyterState = state
 
     @classmethod
-    def init(cls, project,
-             compute_provider,
-             dns_provider,
-             location,
-             dns_id,
-             state_path) -> Union["JupyterController", None]:
+    def init(cls,
+             project: str,
+             compute_provider: str,
+             dns_provider: str,
+             location: str,
+             dns_id: str,
+             state_path: str,
+             ) -> Union["JupyterController", None]:
+        """ it will be deprecated in future releases """
         if not fetch_state(state_path):
             st = JupyterState(
                 project=project,
@@ -207,16 +226,40 @@ class JupyterController:
                 self_link=state_path,
             )
             fp = push_state(st)
-            jup = cls.from_state(state_path)
+            compute: ComputeSpec = utils.get_class(
+                VM_PROVIDERS[compute_provider])(keyvar=defaults.JUP_COMPUTE_KEY)
+            dns: DNSSpec = utils.get_class(
+                DNS_PROVIDERS[dns_provider])(
+                    keyvar=_check_dns_env_var())
+            jup = cls(compute, dns=dns, state=st)
             return jup
         return None
 
+    # @classmethod
+    # def from_settings(cls, settings_module: str) -> "JupyterController":
+    #     cfg = load_jupyter_conf(settings_module)
+    #     state = fetch_state(cfg.STATE_PATH)
+    #     compute: ComputeSpec = utils.get_class(
+    #         VM_PROVIDERS[state.compute_provider])(keyvar=defaults.JUP_COMPUTE_KEY)
+    #     dns: DNSSpec = utils.get_class(
+    #         DNS_PROVIDERS[state.dns_provider])(keyvar=_check_dns_env_var())
+
+    #     return cls(
+    #         compute=compute,
+    #         dns=dns,
+    #         state=state
+    #     )
+
     @classmethod
     def from_state(cls, path: str) -> "JupyterController":
+        """ It will be deprecated in future releases"""
         s = fetch_state(path)
-        compute: ProviderSpec = utils.get_class(
-            VM_PROVIDERS[s.compute_provider])()
-        dns: DNSSpec = utils.get_class(DNS_PROVIDERS[s.dns_provider])()
+        compute: ComputeSpec = utils.get_class(
+            VM_PROVIDERS[s.compute_provider])(
+                keyvar=defaults.JUP_COMPUTE_KEY
+        )
+        dns: DNSSpec = utils.get_class(DNS_PROVIDERS[s.dns_provider])(
+            keyvar=defaults.JUP_COMPUTE_KEY)
         return cls(
             compute=compute,
             dns=dns,
@@ -235,33 +278,35 @@ class JupyterController:
     def prj(self) -> str:
         return self._state.project
 
-    def build_url(self, vm_name) -> str:
+    @property
+    def url(self) -> str:
+        return self._state.url
+
+    def _build_url(self, vm_name) -> str:
         url = f"{vm_name}.{self.prj}.{self.zone.domain}"
         return url
 
     def find_node_types(self, ram=2, cpu=2) -> List[InstanceType]:
         if prov.providerid == "gce":
-            nodes = find_gce(self.prov, ram, cpu, gpu)
+            nodes = find_gce(self.compute, ram, cpu, gpu)
         return nodes
 
     def _get_startup_script(self):
         here = os.path.abspath(os.path.dirname(__file__))
-        _file = f"{self.prov.providerid}_startup.sh"
-        # if gpu:
-        #    _file = f"{self.prov.providerid}_startup_gpu.sh"
+        _file = f"{self.compute.providerid}_startup.sh"
         with open(f"{here}/files/{_file}", "r") as f:
             startup = f.read()
         return startup
 
     def import_volume(self, name):
-        _v = self.prov.get_volume(name, location=self.location)
+        _v = self.compute.get_volume(name, location=self.location)
         self._state.volumes[name] = _v
 
     def resize_volume(self, name, size) -> bool:
         vol = self._state.volumes.get(name)
         if vol:
-            res = self.prov.resize_volume(name, size,
-                                          location=self._state.location)
+            res = self.compute.resize_volume(name, size,
+                                             location=self._state.location)
             if res:
                 vol.size = size
                 self._state.volumes[name] = vol
@@ -271,13 +316,13 @@ class JupyterController:
     def destroy_volume(self, name) -> bool:
         vol = self._state.volumes.get(name)
         if vol:
-            res = self.prov.destroy_volume(name, self.location)
+            res = self.compute.destroy_volume(name, self.location)
             del(self._state.volumes[name])
             return res
         return False
 
     def check_volume(self, name) -> bool:
-        vol = self.prov.get_volume(name)
+        vol = self.compute.get_volume(name)
         if vol:
             return True
         return False
@@ -296,29 +341,74 @@ class JupyterController:
             description=description,
             storage_type=storage_type,
         )
-        st = self.prov.create_volume(sr)
+        st = self.compute.create_volume(sr)
         self._state.volumes[name] = st
 
-    def create_lab(self,
-                   container="jupyter/minimal-notebook:python-3.10.6",
-                   uuid="1000",
-                   boot_size="10",
-                   boot_image="debian-11-bullseye-v20220822",
-                   boot_type="pd-standard",
-                   boot_delete=True,
-                   ram=1,
-                   cpu=1,
-                   gpu=None,
-                   registry=None,
-                   network="default",
-                   tags=["http-server", "https-server"],
-                   instance_type=None,
-                   volume_data=None,
-                   lab_timeout=20 * 60,  # in seconds
-                   account: str = None,
-                   roles: List[str] = [],
-                   debug=False
-                   ) -> LabResponse:
+    def create_lab(self, instance: JupyterInstance, *,
+                   volume: Optional[JupyterVolume] = None) \
+            -> LabResponse:
+        """
+        It check for the volumes, create them, create the instance
+        and register the public ip of the instance into the DNS.
+        """
+
+        token = secrets.token_urlsafe(16)
+        if volume:
+            if not self.check_volume(volume.name):
+                self.create_volume(volume.name, size=volume.size)
+                self.push()
+        vm = self.create_instance(**instance.dict(), token=token)
+        record = DNSRecord(
+            name=self.url,
+            zoneid=self.zone.id,
+            record_type="A",
+            data=[
+                vm.public_ips[0]
+            ]
+        )
+        data = self.dns.create_record(record)
+        record.id = data["id"]
+        self._state.record = record
+
+        return LabResponse(
+            url=self.url.strip("."),
+            token=token,
+            project=self.prj
+        )
+
+    def _generate_url(self) -> str:
+        """ it generates a random vm name and creates the url without protocol:
+            lab-<random_string>.<project>.<dns_domain>
+        """
+        _name = utils.generate_random(
+            size=5, alphabet=defaults.NANO_MACHINE_ALPHABET)
+        vm_name = f"lab-{_name}"
+
+        url = f"{vm_name}.{self.prj}.{self.zone.domain}"
+
+        return url
+
+    def create_instance(self,
+                        container="jupyter/minimal-notebook:python-3.10.6",
+                        uuid="1000",
+                        boot_size="10",
+                        boot_image="debian-11-bullseye-v20220822",
+                        boot_type="pd-standard",
+                        boot_delete=True,
+                        ram=1,
+                        cpu=1,
+                        gpu=None,
+                        registry=None,
+                        network="default",
+                        tags=["http-server", "https-server"],
+                        instance_type=None,
+                        volume_data=None,
+                        lab_timeout=20 * 60,  # in seconds
+                        account: str = None,
+                        roles: List[str] = [],
+                        debug=False,
+                        token=None,
+                        ) -> VMInstance:
         if ram and cpu and not instance_type:
             _types = self.find_node_types(ram, cpu)
             if _types:
@@ -326,17 +416,10 @@ class JupyterController:
         else:
             node_type = instance_type
 
-        token = secrets.token_urlsafe(16)
-        _name = utils.generate_random(
-            size=5, alphabet=defaults.NANO_MACHINE_ALPHABET)
 
-        vm_name = f"lab-{_name}"
-        zone = self.zone
-        url = self.build_url(vm_name)
-        self._state.url = url
         to_attach = []
         if volume_data:
-            vol = self.prov.get_volume(volume_data)
+            vol = self.compute.get_volume(volume_data)
             if vol:
                 to_attach = [
                     AttachStorage(
@@ -356,12 +439,16 @@ class JupyterController:
         if account:
             scopes = Permissions(account=account, roles=roles)
 
+        url = self._generate_url()
+        self._state.url = url
+        vm_name = url.split(".")[0]
+
         vm = VMRequest(
             name=vm_name,
             instance_type=node_type,
             startup_script=self._get_startup_script(),
             location=self.location,
-            provider=self.prov.providerid,
+            provider=self.compute.providerid,
             boot=BootDiskRequest(
                 image=boot_image,
                 size=boot_size,
@@ -369,7 +456,7 @@ class JupyterController:
                 auto_delete=boot_delete,
             ),
             metadata={
-                "labdomain": f"{self.prj}.{zone.domain}".strip("."),
+                "labdomain": f"{self.prj}.{self.zone.domain}".strip("."),
                 "laburl": url.strip("."),
                 "labimage": container,
                 "labtoken": token,
@@ -389,30 +476,14 @@ class JupyterController:
             external_ip="ephemeral",
             labels={"project": self.prj},
         )
-        instance = self.prov.create_vm(vm)
+        instance = self.compute.create_vm(vm)
         self._state.vm = instance
-
-        record = DNSRecord(
-            name=url,
-            zoneid=self.zone.id,
-            record_type="A",
-            data=[
-                instance.public_ips[0]
-            ]
-        )
-        data = self.dns.create_record(record)
-        record.id = data["id"]
-        self._state.record = record
-        return LabResponse(
-            url=url.strip("."),
-            token=token,
-            project=self.prj
-        )
+        return instance
 
     def destroy_lab(self):
         if self._state.vm:
-            self.prov.destroy_vm(vm=self._state.vm.vm_name,
-                                 location=self._state.vm.location)
+            self.compute.destroy_vm(vm=self._state.vm.vm_name,
+                                    location=self._state.vm.location)
             self._state.vm = None
         if self._state.record:
             self.dns.delete_record(self.zone.id, self._state.record.id)
@@ -425,7 +496,7 @@ class JupyterController:
         return fp
 
     def fetch(self, console=None):
-        vms = self.prov.list_vms()
+        vms = self.compute.list_vms()
         vm_set = False
         for vm in vms:
             prj = vm.labels.get("project")
@@ -454,7 +525,7 @@ class JupyterController:
                     break
         volumes = {}
         for _vol in self._state.volumes.keys():
-            vol = self.prov.get_volume(_vol)
+            vol = self.compute.get_volume(_vol)
             if vol:
                 volumes[_vol] = vol
         self._state.volumes = volumes
@@ -464,3 +535,62 @@ class JupyterController:
 
     def clean(self):
         clean_state(self._state.self_link)
+
+
+def save_conf(state_path, conf_path=defaults.JUPCTL_CONF):
+    utils.write_toml(conf_path, {"state": state_path})
+
+
+def _load_state_path(conf_path=defaults.JUPCTL_CONF) -> str:
+    data = utils.read_toml(conf_path)
+    return data.get("state")
+
+
+def from_state(path: str) -> JupyterController:
+    """ It creates JupyterController from the state """
+    s = fetch_state(path)
+    compute: ComputeSpec = utils.get_class(
+        VM_PROVIDERS[s.compute_provider])(
+            keyvar=defaults.JUP_COMPUTE_KEY
+    )
+    dns: DNSSpec = utils.get_class(DNS_PROVIDERS[s.dns_provider])(
+        keyvar=defaults.JUP_COMPUTE_KEY)
+    return JupyterController(
+        compute=compute,
+        dns=dns,
+        state=s
+    )
+
+
+def from_conf(conf_path: str = defaults.JUPCTL_CONF) -> JupyterController:
+    _state = _load_state_path(conf_path)
+    jup = from_state(_state)
+    return jup
+
+
+def init(project: str,
+         compute_provider: str,
+         dns_provider: str,
+         location: str,
+         dns_id: str,
+         state_path: str,
+         ) -> Union[JupyterController, None]:
+    """ it will be deprecated in future releases """
+    if not fetch_state(state_path):
+        st = JupyterState(
+            project=project,
+            compute_provider=compute_provider,
+            dns_provider=dns_provider,
+            location=location,
+            zone_id=dns_id,
+            self_link=state_path,
+        )
+        fp = push_state(st)
+        compute: ComputeSpec = utils.get_class(
+            VM_PROVIDERS[compute_provider])(keyvar=defaults.JUP_COMPUTE_KEY)
+        dns: DNSSpec = utils.get_class(
+            DNS_PROVIDERS[dns_provider])(
+                keyvar=_check_dns_env_var())
+        jup = JupyterController(compute, dns=dns, state=st)
+        return jup
+    return None
