@@ -1,11 +1,20 @@
 import json
 import os
 from typing import Any, Dict, Optional
-from labmachine import defaults
+
 from pydantic import BaseSettings
 
 
+GOOGLE_AUTH_ENV = "GOOGLE_APPLICATION_CREDENTIALS"
+
 class GCConf(BaseSettings):
+    """
+    :param CREDENTIALS: credential file open by python
+    :param PROJECT: projectid
+    :param LOCATION: region/zone
+    :param SERVICE_ACCOUNT: client_email
+
+    """
     CREDENTIALS: str
     PROJECT: str
     LOCATION: Optional[str] = None
@@ -15,8 +24,11 @@ class GCConf(BaseSettings):
         env_prefix = "GCE_"
 
 
-def get_auth_conf() -> GCConf:
-    creds_env = os.environ.get("GOOGLE_APPLICATION_CREDENTIALS")
+def get_auth_conf(env_var=GOOGLE_AUTH_ENV) -> GCConf:
+    """
+    https://googleapis.dev/python/google-api-core/latest/auth.html#authentication
+    """
+    creds_env = os.environ.get(env_var)
     if creds_env:
         with open(creds_env, "r") as f:
             data = json.loads(f.read())
